source "./filterFunction.tcl"
source "./filterLeastSquares.tcl"
source "./Common functions.tcl"
source "./P100GUtils.tcl"
source "./NoiseGeneration.tcl"

# Pi is used in a couple of places. Make it global
global pi
set pi 3.1415926535897931

# ########################################################################################################
# Description of filenames and locationos
#
# At each stage of the test, various files are created and used.
# The names of folders and files are determined either in the sequence object or the step object
# (depending on whether the scope of the folder/file is the entire sequence or a single step).
# The location and names of files is dependent on whether the GUI is running on a remote host
# There are also some couplings with demo mode and reprocessing.
#
# Profiles Folder: (sequence)
#	Local: 		Current folder (where the script is running)
#	Remote: 	Passed in as an argument to the script
# Profiles File: (step)
#				Held in stimulusList
# Logs Folder: (sequence)
# 				[pwd]/Logs
# Log File: (sequence)
#	Normal: 	Log <noiseTransferType> <dateStart>-T-<timeStart>.txt
#	Demo: 		Log <noiseTransferType> Demo.txt
# Results Folder: (sequence)
#   Normal: 	[pwd]/Results/Results <noiseTransferType> <dateStart>-T-<timeStart>
#	Demo: 		[pwd]/Results/Results <noiseTransferType> Demo
# Overall Results File: (sequence)
# 	Normal: 	<Results Folder>/Results <dateStart>-T-<timeStart>.csv
# 	Demo: 		<Results Folder>/Results Demo.csv
# Overall Results GUI Picture: (sequence)
#				<Results Folder>/Results.bmp
#
# The test process is essentially:
# 	Capture (save to file)
#	Load into CAT
#	Export from CAT (into file)
#	Process file (results saved to file)
#	Individual step results are stored in the voerall results file above
#
# The folders and files used at each stage above are:
#
# Capture Folder: (step)
#	Normal: 			<Results Folder> (as above)
#	Demo, not Remote: 	[pwd]/ReProcess
#	Remote: 			Passed in as an argument to the script
# Capture File: (step)
#						<noiseTransferType> NoiseTransfer <freq>Hz.clxz
# CAT Export Folder: (step)
#	Local:				<Results Folder>
#	Remote:				Defaults to Calnex/.../CAT/Export
# CAT Export File: (step)	
#						NoiseTransfer_<freqWithUnderbar>_<meas>_Cat.csv		where <meas> is T4 or 1PPS etc
# Filtered CAT Data Folder: (step)
#						<Results Folder>
# Filtered CAT Data: (step)
#						NoiseTransfer_<freqWithUnderbar>_Results.csv	
# 

# #######################################################################################################
#   STIMULUS 
#	Either PTP or SyncE will be applied as a noise stimulus.
#   The stimulus applied is defined in stimulusList.
#
#   PTP:   Freq, pk-pk ampl, packet sample rate, capture duration, lower limit, upper limit, PDV file
#   SYNCE: Freq, pk-pk ampl, packet sample rate, capture duration, lower limit, upper limit
#
#	The frequencies, amplitudes and limits are based on those defined in amendment TD64 Rev2 to G.8273.2
#	The SyncE noise amplitudes at lower frequencies are higher than those suggested in the standard.
#	Using higher amplitudes provides better signal-to-noise ratio whilst being within the tolerable levels for the DUT.
#	Note that the limits are defined in Table VI.4 (PTP) and VI.5 (SyncE).
#	Since the script uses the least squares filter, the limits are defined with N=10 for PTP and N=25 for SyncE
#	The limits are checked against the measured pk-pk amplitude (as opposed to the gain)
#
#	NOTES: 
#	1. 	The amplitude of the noise applied for a PTP input is defined in the standards as 200ns.
#		The profiles loaded by P-X are however applied in only one direction (to Syncs).
#		So, to achieve 200ns 2Way noise, 400ns is applied to the Sync messages.
#		In other words, the profiles used have a pk-pk amplitude of 400ns.
#		The table below has been changed to reflect the fact that what we are actually applying is
#		2Way noise of 200ns (it used to be 400ns).
#	2.	The sample rate and amplitudes must be represented as floating point to make sure subsequent 
#		maths operations are not rounded to integers. 
# #######################################################################################################
#
# When stimulus is PTP
# Limits below are from Appendix VI in the standard using N=10
set stimulusListPtp { 
	{ 0.00390625 200.0 16.0 800 130	215 "NoiseTransfer_0_00390625Hz.csv" }
	{ 0.0078125  200.0 16.0 800 130	215 "NoiseTransfer_0_0078125Hz.csv"  }
	{ 0.015625   200.0 16.0 400 130	215 "NoiseTransfer_0_015625Hz.csv"   }
	{ 0.03125    200.0 16.0 250 130	215 "NoiseTransfer_0_03125Hz.csv"    }
	{ 0.0615625  200.0 16.0 250 NA  215 "NoiseTransfer_0_0615625Hz.csv"  }
	{ 0.123125   200.0 16.0 250 NA  140 "NoiseTransfer_0_123125Hz.csv"   }
	{ 0.24625    200.0 16.0 250 NA  90  "NoiseTransfer_0_24625Hz.csv"    }
	{ 0.4925     200.0 16.0 250 NA  50  "NoiseTransfer_0_4925Hz.csv"     }
	{ 0.985      200.0 16.0 250 NA  35  "NoiseTransfer_0_985Hz.csv"      }
	{ 1.985      200.0 16.0 250 NA  30  "NoiseTransfer_1_985Hz.csv"      }
}
# When stimulus is SyncE
#set Nsynce 25
# OLD VALUES, NOW REPLACED. Need to keep the old list for re-processing
set stimulusListSyncE_old {
	{ 0.00391 250.0 16.0 800 NA  45 } 
	{ 0.00781 250.0 16.0 800 NA  65 }
	{ 0.01563 250.0 16.0 400 NA  100 }
	{ 0.03125 250.0 16.0 250 NA  160 } 
	{ 0.06156 250.0 16.0 250 NA  285 } 
	{ 0.12313 250.0 16.0 250 150 285 } 
	{ 0.24625 250.0 16.0 250 150 285 } 
	{ 0.4925  250.0 16.0 250 150 285 } 
	{ 0.985   250.0 16.0 250 150 285 } 
	{ 1.985   250.0 16.0 250 NA  285 } 
	{ 3.985   250.0 16.0 250 NA  285 } 
	{ 7.985   250.0 16.0 250 NA  285 } 
}
set stimulusListSyncE {
	{ 0.00391 2000.0 16.0 800 NA  225 } 
	{ 0.00781 2000.0 16.0 800 NA  340 }
	{ 0.01563 2000.0 16.0 400 NA  630 }
	{ 0.03125 1000.0 16.0 250 NA  565 } 
	{ 0.06156 500.0 16.0 250 NA  545 } 
	{ 0.12313 250.0 16.0 250 150 285 } 
	{ 0.24625 250.0 16.0 250 150 285 } 
	{ 0.4925  250.0 16.0 250 150 285 } 
	{ 0.985   250.0 16.0 250 150 285 } 
	{ 1.985   250.0 16.0 250 NA  285 } 
	{ 3.985   250.0 16.0 250 NA  285 } 
	{ 7.985   250.0 16.0 250 NA  285 } 
}

# P100G analysis is a bit of a hack. Some vars to support it...
#global P100GNoiseTransferFile
#set P100GNoiseTransferFile TRUE
# T4 or 1pps
#set P100GMeas "1pps"
#set P100GStimulus "SyncE"

proc initialiseLogging {demoMode noiseTransferType} {
	set folderLogs "Logs"
	set dt [currentDateTime]
	set d [lindex $dt 0]
	set t [lindex $dt 1]

	if {!$demoMode} {
		set fileLog "$folderLogs/Log $noiseTransferType $d-T-$t.txt"
	} else {
		set fileLog "$folderLogs/Log $noiseTransferType Demo.txt"
	}

	openLogFile $folderLogs $fileLog
	log "Log will be stored in:\t\"$fileLog\""

}


proc noiseTransferInitialise {paragon type parameters scriptVersion demoMode remoteHost} {
	variable P100GNoiseTransferFile
	variable stimulusListPtp
	variable stimulusListSyncE
	variable stimulusListSyncE_old

	# If we are processing a P100G file, then load in the code to extract deltails from the header
	if {$P100GNoiseTransferFile} { source "P100GCapture.tcl" }
	
	transferSequence create sequence $type $parameters $demoMode $remoteHost

	set stimulus [sequence getStimulus]
	
	# A setting allows the user to run only a sub-set of points
	# Get the setting and construct the stimulusList appropriately
	set syncEPoints [dict get $parameters "SyncEPoints"]
	set ptpPoints [dict get $parameters "PtpPoints"]
	set reProcess [dict get $parameters "ReProcess"]
	set simulatedData [dict get $parameters "SimulatedData"]
	
	if {$reProcess == "TRUE"} {
		# Read in the stimulusList.tcl file generated when originally running the test
		# This file needs to have been copied into the ReProcess folder
		# If the file does not exist, then assume that the OLD list was used
		# NOTE: Need to figure out what to do with P100G here...
		set slFile "./ReProcess/stimulusList_$stimulus.tcl"
		if {[file exists $slFile]} {
			XXXDebugTrace "stimulusList read from $slFile"
			source "$slFile"
		} else {
			XXXDebugTrace "$slFile not found. Using hard-coded version"
			if {$stimulus == "PTP"} {set stimulusListUsedDuringTest $stimulusListPtp } \
			elseif {$stimulus == "SYNCE"} {set stimulusListUsedDuringTest $stimulusListSyncE_old} \
			else {
				logLevel "FATAL" "unrecognised stimulus - $stimulus"
			}
		}
	}
	
	if {$stimulus == "PTP"} {
		set points $ptpPoints
		set stimulusList $stimulusListPtp
	} elseif {$stimulus == "SYNCE"} {
		set points $syncEPoints
		set stimulusList $stimulusListSyncE
	} else {
		logLevel "FATAL" "unrecognised stimulus - $stimulus"
	}

	# Parse the P100G file to extract the stimulus list
	if {$P100GNoiseTransferFile == TRUE} {
		set stimulusList [P100G_parseCaptureFile sequence $stimulusList]
	}
	
	if {$reProcess == "FALSE"} {outputStimulusListAsTcl sequence $stimulusList}
	
	set selectedStimulusList {}
	if {$points != "ALL"} {
		foreach p $points {
			lappend selectedStimulusList [lindex $stimulusList $p]
		}
		set stimulusList $selectedStimulusList
	}
	
	if {$reProcess == "TRUE" && $simulatedData == "FALSE"} {set stimulusList $stimulusListUsedDuringTest}
	if {$reProcess == "TRUE" && $simulatedData != "FALSE"} {set stimulusList $selectedStimulusList}
	
	sequence loadStepList $stimulusList
	#sequence outputStepListAsTcl
	
	log  "*************************************"
	log  "Enhanced Noise Transfer\t$scriptVersion"
	log  "*************************************"
	log	 ""

	sequence printParameters
	
	set serialNumber [$paragon getInstrumentSerialNumber]
	set version [$paragon getInstrumentVersion]
	
	###############################################################################
	# Configure graphing
	#------------------------------------------------------------------------------
	# Only include graphing if required - needs a couple of packages installed
	set g [dict get $parameters "Graphing"]
	if {$g == "TRUE" || $g == "True"} {
		XXXDebugTrace "Including graphing.tcl"
		source "./graphing.tcl"
		# Graph initialisation is done later for P100G
		if {$P100GNoiseTransferFile != TRUE} {
			configureGraphing [sequence getStimulus] [sequence getMeasurement] [sequence getStepList] $parameters $serialNumber $version [sequence getStartDate] [sequence getStartTime]
		} else {
			set meas [sequence getMeasurement]
			if {$meas == "PTP" || $meas == "BOTH"} {
				set capture [sequence p100G_getPtpCapture]
			}
			if {$meas == "1PPS"} {
				set capture [sequence p100G_get1ppsCapture]				
			}
			configureGraphing [sequence getStimulus] [sequence getMeasurement] [sequence getStepList] $parameters [$capture getSetting serial] [$capture getSetting softwareVersion] [$capture getSetting startTime] [$capture getSetting startTime]
		}
	}
}

proc outputStimulusListAsTcl {seq sList} {
		set stim [$seq getStimulus]
		set folder [$seq getFolderResults]
		set sListFilename "stimulusList_$stim.tcl"
		set sListPath "$folder/$sListFilename"
		
		# If the file already exists, delete it
		if {[file exists $sListPath]} {
			file delete -force $sListPath
			log "Existing stimulus list file \"$sListPath\" has been deleted"
		}

		set fid [open $sListPath "a"]
		
		puts $fid "# THIS FILE HAS BEEN GENERATED BY THE G.8273.2 SCRIPT. DO NOT DELETE OR MODIFY"
		puts $fid "set stimulusListUsedDuringTest \{ "
		foreach tp $sList {
			puts -nonewline $fid "\{ "
			puts -nonewline $fid "[lindex $tp 0] "
			puts -nonewline $fid "[lindex $tp 1] "
			puts -nonewline $fid "[lindex $tp 2] "
			puts -nonewline $fid "[lindex $tp 3] "
			puts -nonewline $fid "[lindex $tp 4] "
			puts -nonewline $fid "[lindex $tp 5] "
			puts $fid "[lindex $tp 6] \}"
		}
		puts $fid " \}"
		flush $fid
		close $fid

}

proc doBaselineNoiseGenerationMeasurement {measurement isDemo simulatedData} {	
	# Capture duration is set here but should maybe be set at a higher level
	# Packet rate is used to calculate the number of packet we expect in the capture
	# This should probably come from a higher level but right now, hard-code here
	set captureDuration 120
	set packetRate 16
	# The noiseGeneration function takes a dictionary as input
	# This specifies the capture time and the desired measurements
	dict set noiseGenerationDict "captureDuration" $captureDuration
	# We are running at 16pkts/second and so we would expect 20*16 messages
	# Allow for 10% slop
	set expectedMessageCount [expr {$captureDuration*$packetRate}]
	set expectedMessageThreshold [expr {$expectedMessageCount - ($expectedMessageCount*0.2)}]
	if {$measurement == "PTP" || $measurement == "BOTH"} {
		set noiseGenerationPtp {T4 range T4 revMessages}
		dict set noiseGenerationMeasDict "PTP" $noiseGenerationPtp		
	}
	if {$measurement == "1PPS" || $measurement == "BOTH"} {
		set noiseGeneration1pps {1PPS range 1PPS min 1PPS max}
		dict set noiseGenerationMeasDict "1PPS" $noiseGeneration1pps		
	}
	dict set noiseGenerationDict "measurements" $noiseGenerationMeasDict
	#puts $noiseGenerationDict
	
	# The noiseGeneration function returns a dictionary containing the requested results
	#set resultsDict [noiseGeneration par1 "[pwd]//[sequence getFolderResults]" "BaselineNoiseGeneration.clxz" $noiseGenerationDict $isDemo "[pwd]//ReProcess/"]
	set resultsDict [noiseGeneration par1 "[pwd]/[sequence getFolderResults]" "baselineNoiseGeneration.clxz" $noiseGenerationDict $isDemo "[pwd]/ReProcess/"]
	XXXDebugTrace "Baseline measurements - $resultsDict"

	if {$simulatedData == "FALSE"} {
		if {$measurement == "PTP" || $measurement == "BOTH"} {
			set revMessages [dict get $resultsDict "T4revMessages"]
			if {$revMessages < $expectedMessageThreshold && ($measurement == "PTP" || $measurement == "BOTH")} {
				logLevel "FATAL" "The are $revMessages Del-Req/Del-Resp messages in the capture\n \
				There should be approximately $expectedMessageCount messages\n \
				Please check your configuration manually before running this script"
			}
		}
		if {$measurement == "1PPS" || $measurement == "BOTH"} {
			set 1ppsRange [dict get $resultsDict "1PPSrange"]
			set 1ppsMin [dict get $resultsDict "1PPSmin"]
			set 1ppsMax [dict get $resultsDict "1PPSmax"]
			if {$1ppsRange < 0.5 && $1ppsMin < 0.5 && $1ppsMin > -0.5} {
				logLevel "WARNING" "The 1pps baseline measurement is questionable\nThis may not be an issue but you should check your results and configuration"
			}
		}
	} else {
		log "Simulated data - not checking baseline noise measurement..."
	}
}

#ENT-41  - add noiseTransferType as a param

proc timeNoiseTransfer {par1 noiseTransferType} {
	variable demoMode
	variable simulatedData
	variable P100GNoiseTransferFile
	variable P100GMeas
	variable P100GStimulus
	variable graphing

	set stimulus [sequence getStimulus]
	set measurement [sequence getMeasurement]
	
	if (!$demoMode) {
		par1 paragonset MasterSlave Impairments ProfileReplay MasterTx Corruption Enable FALSE
		par1 paragonset MasterSlave Impairments ProfileReplay MasterRx Corruption Enable FALSE

		if {$stimulus == "PTP"} {
			# If PTP is the stimulus, then enable delay impairments
			par1 paragonset MasterSlave Impairments VariableDelay MasterTx Enable TRUE
		}
	}

	# If we are processing a P100G capture, it contains all the table steps
	# So, load the cdf into the CAT and export the data.
	# As we loop through each point in the table, the relevant section from
	# the capture will be used.
	# Note that the PTP and 1PPS sections here are effectively copies of each other
	# This could be substantially re-factored 
	if {$P100GNoiseTransferFile} {
		#source "P100GCapture.tcl"
		
		# Get the capture file 
		# If the measurement is PTP, then look in ReprocessP100G for Ptp1588100.cdf
		# If the measurement is 1pps, then look in ReporcessP100G for OnePpsAccuracy100.cdf
		# If the measaurement is BOTH, then use both of the above.
		if {$measurement == "PTP" || $measurement == "BOTH"} {
			set measure "PTP"

			set fullResults [P100G_getFullResults sequence $measure]
			P100G_processSequence sequence $fullResults

			#set sList [sequence getStepList]
			###set capture [sequence p100G_getPtpCapture]
			# Dirty hack for testing - override the values in the capture. Need to figure out how to deal with these properly		
			if {$graphing} {
				savePlotWindow "P100G/" Results
			}		
		}
		if {$measurement == "1PPS" || $measurement == "BOTH"} {    
			set measure "1PPS"

			set fullResults [P100G_getFullResults sequence $measure]
			P100G_processSequence sequence $fullResults

			# For testing - override the values in the capture
			#set deviceRecoveryTime 5 
			#set guardTime 15

			if {$measurement != "BOTH"} {
				#configureGraphing $stimulus $measurement $sList "" [$capture getSetting serial] [$capture getSetting softwareVersion] [$capture getSetting startTime] [$capture getSetting startTime]
			}
			
			if {$graphing} {
				savePlotWindow "P100G/" Results
			}
			
			#capture destroy
			return
		}
		
	}		
		
	set counter 0

#
#    Start MSE first then wait on any device settling Time
#    ** DO NOT START MSE if in DEMO mode
#

    if {!$demoMode && $simulatedData == "FALSE"} {
		# select packet generation with SSM code of QL_PRC
		par1 paragonset PacketGeneration #0 Esmc #0 SsmType QL-PRC
		par1 paragonset PacketGeneration #0 Enable TRUE

		set captureFilename [sequence getFolderResults]
		append captureFilename "/settlingTime.clxz"

# ENT-41  pass in $noiseTransferType
		startMSEandWaitonSettling  par1 $noiseTransferType [sequence getSettlingTime] $captureFilename $demoMode 
	}

	if {$simulatedData == "FALSE"} {
		log "\n*** Baseline Noise Generation Measurement ***"
		doBaselineNoiseGenerationMeasurement $measurement $demoMode $simulatedData
	}	
	
	log "\n*** Stimulus = $stimulus ---> DUT ---> Measure = $measurement ***"	
	set seq [sequence getSequence]
	#log $seq
	foreach seqStep $seq {
		set i [lindex $seqStep 0]
		set step [lindex $seqStep 1]

		set inputFreq [$step getFreq]
		set inputAmpl [$step getAmpl]
		set inputSampleRate [$step getSampleRate]
		set captureTime [$step getDuration]
		set lowerLimit [$step getLL]
		set upperLimit [$step getUL]

		log ""
		log "[currentDateTime]\tStep $i Stimulus:\t$inputFreq\t$inputAmpl\t$captureTime"

		set tp [sequence getStepList]
		
		$step startStimulus
		$step doCapture
		$step stopStimulus

		if {$measurement == "PTP" || $measurement == "BOTH"} {
			set measure "PTP"
			XXXDebugTrace "Opening CAT (PTP)..."
			openCat par1 PTP

			$step exportFileFromCat "PTP"

			# Read the results (either from file or from simulated data)
			# This also removes the recovery time
			$step getResults "PTP"
			
			$step applyFilter "PTP"
			$step calculateFinalResults "PTP"
			if {$measurement != "BOTH"} {
				$step saveAndDisplayFinalResults "PTP"
			}
			log "[currentDateTime]\tStep $counter $measure Meas:\t[format %10.8f [$step getFreq]]\t [$step getAmpl]\t [format %.2f [$step getPtpAmplitudeOut]] \t [format %.2f [$step getPtpGain]]\t [$step getPtpPassFail]"
		
		} 
		
		if {$measurement == "1PPS" || $measurement == "BOTH"} {    
			set measure "1PPS"
			XXXDebugTrace "Opening CAT (1pps)..."
			openCat par1 1PPS

			$step exportFileFromCat "1PPS"
			$step getResults "1PPS"
			$step applyFilter "1PPS"
			$step calculateFinalResults "1PPS"
			if {$measurement != "BOTH"} {
				$step saveAndDisplayFinalResults "1PPS"
			}
			log "[currentDateTime]\tStep $counter $measure Meas:\t[format %10.8f [$step getFreq]]\t [$step getAmpl]\t [format %.2f [$step get1ppsAmplitudeOut]] \t [format %.2f [$step get1ppsGain]]\t [$step get1ppsPassFail]"


			par1 clearDictionaryMeasurements
		} 

		if {$measurement == "BOTH"} {
			$step saveAndDisplayFinalResults "BOTH"
		}

		incr counter
    }

#
#    FINISHED - so stop  MSE  
#    ENT50 - only need to do this if not reprocessing data

    if {!$demoMode} {
		stopMSE par1 $captureTime 
	}
	
	if {$graphing} {
		set folder [sequence getFolderResults]
		savePlotWindow $folder Results
	}
	
	log "\nTest Complete"
	sequence destroy
}

# ############################################################################################################################################################
# Functions to save results to file and to update the UI
proc saveResultsToFile { openedResultsGeneralFile f frequency pkpkAmplitudeIn ptpAmplitudeOut ptpGain 1ppsAmplitudeOut 1ppsGain ll ul ptpPass 1ppsPass} {
	variable graphing
	
	set resultsInfo ""
	set file [ string map {".csv" ""} $f ]

	append resultsInfo $file ", " $frequency ", " $pkpkAmplitudeIn ", " [format %.2f $ptpAmplitudeOut] ", " [format %.2f $ptpGain] ", " [format %.2f $1ppsAmplitudeOut] ", " [format %.2f $1ppsGain] ", " $ll ", " $ul ", " $ptpPass ", " $1ppsPass
	if {$graphing} {
		updatePtpBodePlot $frequency $ptpAmplitudeOut $ptpGain $ptpPass
		update1ppsBodePlot $frequency $1ppsAmplitudeOut $1ppsGain $1ppsPass
	}
	puts $openedResultsGeneralFile $resultsInfo
}

proc savePtpResultsToFile { openedResultsGeneralFile f frequency pkpkAmplitudeIn ptpAmplitudeOut ptpGain ll ul ptpPass} {
	variable graphing
	
	set resultsInfo ""
	set file [ string map {".csv" ""} $f ]

	append resultsInfo $file ", " $frequency ", " $pkpkAmplitudeIn ", " [format %.2f $ptpAmplitudeOut] ", " [format %.2f $ptpGain] ", " $ll ", " $ul ", " $ptpPass
	if {$graphing} {updatePtpBodePlot $frequency $ptpAmplitudeOut $ptpGain $ptpPass}
	puts $openedResultsGeneralFile $resultsInfo
}

proc save1ppsResultsToFile { openedResultsGeneralFile f frequency pkpkAmplitudeIn 1ppsAmplitudeOut 1ppsGain ll ul 1ppsPass} {
	variable graphing
	
	set resultsInfo ""
	set file [ string map {".csv" ""} $f ]

	append resultsInfo $file ", " $frequency ", " $pkpkAmplitudeIn ", " [format %.2f $1ppsAmplitudeOut] ", " [format %.2f $1ppsGain] ", " $ll ", " $ul ", " $1ppsPass
	XXXDebugTrace "$resultsInfo"
	if {$graphing} {update1ppsBodePlot $frequency $1ppsAmplitudeOut $1ppsGain $1ppsPass}
	puts $openedResultsGeneralFile $resultsInfo
}
# ############################################################################################################################################################

# ############################################################################################################################################################
# Functions to read results from files exported from the CAT

# --------------------------------------------------------------------------------
# Read results from the CAT export
# ---------------------------------------------------------------------------------
# This reads the test results from the csv file exported from the Cat
# The file contains 2 columns: timestamps, data
#
# Input Parameters
# - catFile: 	 	The file name
# - measure:		PTP or 1PPS
# Output
# - data: 		List containing 2 lists. The first is a list of timestamps;
#					the second is a list of amplitude values.
#					For PTP, the timestamps come from the file
#					For 1PPS, the timestamps are generated - one per second
proc readResultsFromLocalFile {catFile measure freq} {
	set results ""
	XXXDebugTrace "$catFile $measure"

	if {$measure == "PTP"} {
		set results [readPtpResultsFromLocalFile $catFile $freq]
	} elseif {$measure == "1PPS"} {
		set results [read1ppsResultsFromLocalFile $catFile $freq]
	} else {
		log "ERROR: $measure is not a valid measurement type. Aborting."
		exit
	}

	if {([llength $results] != 2) || ([llength [lindex $results 0]] <= 1)} {
		log "FATAL ERROR (reeadResultsFromLocalFile): "
		log "\tThere are insufficient points in the data. Check that your instrument and DUT configuration is correct."
		log "\tYou should verify your set up by running a manual test before re-running the script."
		log "Aborting..."
		exit
	}

	return $results
}	

proc readPtpResultsFromLocalFile {catFilename freq} {
	variable graphing
	XXXDebugTrace "$catFilename"

	set catFile [open $catFilename]
	set x 0
	set y {}
	set times {}
	set data {}
	set index 0
	set dataFound 0

	while {1} {
		# Check for an empty file...
		if {[catch {
				set line [gets $catFile]
				#puts $line
				set line [split $line ";"]
				set timeColumnValue [lindex $line 0]
				set dataColumnValue [lindex $line 1]			
			}]} {
				log "FATAL ERROR (reeadPtpResultsFromLocalFile): "
				log "\tProblem reading $catFile"
				log "\tCheck that your instrument and DUT configuration is correct."
				log "\tYou should verify your set up by running a manual test before re-running the script."
				log "Aborting..."
				close $catFile
				exit
		}

		#puts "$index - t:$timeColumnValue d:$dataColumnValue"
		# Only checking the data column to look for the end of the file. Not sure if checking the timestamp column is necessary.
		if { $dataColumnValue == ""  || $dataColumnValue == "DATA END" } {
			close $catFile
			break
		}
		
		# check token for start of data
            
        if { $dataColumnValue == "value" && $timeColumnValue =="timestamp" } {
			incr dataFound
		}
            
        
		
		# Skip the header - this should be smarter. Maybe look for numeric value in timestamp?
		if {$dataFound > 1 } {
			#dict lappend y $timeColumnValue $dataColumnValue
			lappend times [expr {$timeColumnValue / 1e9}]
			lappend data $dataColumnValue

			incr index
		} 
		incr x
		if { $dataFound == 1 } {
			incr dataFound
		}
		
	}

	# y now contains the the timestamp and measured amplitude in ns for each sampled point
	lappend y $times $data
	if {$graphing} {
		updateDataPlot $times $data $freq "PTP"
	}

	return $y
}

proc read1ppsResultsFromLocalFile {catFilename freq} {
	variable graphing
	XXXDebugTrace "$catFilename"

	set catFile [open $catFilename]
	set x 0
	set y {}
	set times {}
	set data {}
	set index 0

	while {1} {
		if { [catch {set line [gets $catFile]}] } {
			log "FATAL ERROR (read1ppsResultsFromLocalFile): "
			log "\tProblem reading $catFile"
			log "\tCheck that your instrument and DUT configuration is correct."
			log "\tYou should verify your set up by running a manual test before re-running the script."
			log "Aborting..."
			close $catFile
			exit			
		}
		#puts $line
		#set line [split $line ";"]
		#set timeColumnValue [lindex $line 0]
		set dataColumnValue $line
		#puts "$index - t:$timeColumnValue d:$dataColumnValue"
		# Check the data column to look for the end of the file.
		if { $dataColumnValue == ""  || $dataColumnValue == "DATA END" } {
			close $catFile
			break
		}
		# Skip the header - this should be smarter. Maybe look for numeric value in timestamp?
		if {$x > 5} {
			#dict lappend y $timeColumnValue $dataColumnValue
			#lappend times [expr $timeColumnValue / 1e9]
			lappend data $dataColumnValue

			incr index
			lappend times $index
		} 
		incr x
	}

	# y now contains the the timestamp and measured amplitude in ns for each sampled point
	lappend y $times $data
	if {$graphing} {
		updateDataPlot $times $data $freq "1pps"
	}

	return $y
}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Remove recovery and guard time
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Recovery Time: 
#%% When a stimulus is intially applied to the DUT, it may take some time to
#%% recover. So, this function removes a section of the data so it is not taken
#%% into account in the results.
#%%
#%% Guard Time:
#%% For P100G captues, there is a guard time at the end that is removed by the 
#%% CAT. So, this function removes a section of the data at the end so that we have
#%% a better comaprre with P100G results.
#%%
#%% Note: The section of data removed may not match exactly to the time since
#%%       the actual sample rate may not match the expected sample rate.
#%%       This is OK since we don't need to be exact.
#%%
#%% Input Parameters
#%% - results: 	 	The data. Two lists; the first is timestamps; the second is the data
#%% - t:      		The time (in seconds) to remove
#%% - f: 			The sample rate
#%% Output
#%% - testData: 	List containing 2 lists. The first is the adjusted list of timestamps;
#%%					the second is the adjusted list of amplitude values
#%%
proc removeRecoveryTime {results t f freq meas} {
	variable graphing
	
	set range {}
	#set numSamples [expr {int([expr {$t*$f}])}]
	set numSamples [expr {int($t*$f)}]

	set times [lindex $results 0]
	set data [lindex $results 1]
	XXXDebugTrace "($t $f) - $numSamples samples from [llength $data]"

	if {[llength $data] <= $numSamples} {
		logLevel "FATAL" \
		"There are insufficient points in the data. Check that your instrument and DUT configuration is correct.\n/
		You should verify your set up by running a manual test before re-running the script."
	}
	#set range [lrange $results $numSamples end]
	set timeRange [lrange $times $numSamples end]
	set dataRange [lrange $data $numSamples end]
	XXXDebugTrace "($t $f) - Now have [llength $dataRange] samples"

	#puts $range
	lappend range $timeRange $dataRange
	if {$graphing} {
		updateDataPlot $timeRange $dataRange $freq $meas
	}
	return $range
}
proc removeGuardTime {results t f freq meas} {
	variable graphing
	
	set range {}
	set numSamples [expr {int($t*$f)}]
	XXXDebugTrace "($t $f) - $numSamples samples"

	set times [lindex $results 0]
	set data [lindex $results 1]

	if {[llength $data] <= $numSamples} {
		logLevel "FATAL" /
		"There are insufficient points in the data. Check that your instrument and DUT configuration is correct.\n/
		You should verify your set up by running a manual test before re-running the script."
	}

	#set range [lrange $results $numSamples end]
	set timeRange [lrange $times 0 end-$numSamples]
	set dataRange [lrange $data 0 end-$numSamples]

	#puts $range
	lappend range $timeRange $dataRange
	if {$graphing} {
		updateDataPlot $timeRange $dataRange $freq $meas
	}
	return $range
}


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Test data generation
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% This generates data to use as simulated input. 
#%% In real-life, this data will be the 2Way Time Error measured by the CAT
#%%
#%% Input Parameters
#%% - Fsample: 	sample rate
#%% - freq:      Frequency of the generated data
#%% - amplitude: Amplitude of the generated data
#%% - duration: 	period over which the data is generated
#%% Output
#%% - testData: 	List containing 2 lists. The first is a list of timestamps;
#%%					the second is a list of amplitude values
#%%
proc generateSimulatedResults {Fsample freq amplitude duration} {
	variable pi
	variable graphing
	XXXDebugTrace "$Fsample $freq $amplitude $duration"

	set times {}
	set data {}
	set y {}
	set sigma 6
	set N [expr {$Fsample * $duration}]

	# This function outputs +/-amplitude whereas we want pk-pk amplitude so half it
	set ampl [expr {$amplitude / 2}]
	
	for {set i 0} {$i < $N} {incr i} {
		set t [expr {$i/$Fsample}]
		set x [expr {$ampl * sin(2*$pi*$freq*$t) + 5000}]
		set u [expr {rand()*$sigma}]
		set v [expr {$x+$u}]
		lappend times $t
		lappend data $v
		#lappend y [expr $x+$u]
	}
	#return [array get y]
	lappend y $times $data
	if {$graphing} {
		updateDataPlot $times $data $freq "PTP"
	}
	
	return $y
}
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Test data generation
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% This generates data to use as simulated input. 
#%% In real-life, this data will be the 2Way Time Error measured by the CAT
#%%
#%% Input Parameters
#%% - Fsample: 	sample rate
#%% - freq:      Frequency of the generated data
#%% - amplitude: Amplitude of the generated data
#%% - duration: 	period over which the data is generated
#%% Output
#%% - testData: array of amplitude values
#function testData = generateTestData (Fsample, freq, amplitude, duration)
#  sigma	=	6;		% Gaussian noise STD
#	N = Fsample * duration;
#	t = (1:N)'/Fsample;
#	x =	amplitude*sin(2*pi*freq*t+0.12345678);	% Add a constant "unknown" phase for repeatability of the simulation
#	u =	randn(size(x))*sigma;
#	y =	x+u;
#	testData = y;
#endfunction


# ##############################################################################
# Create the folder structure for recording results
# ------------------------------------------------------------------------------
#proc createResultsFolders {folderResults folderTestRunResults} {
#	if {![file exists $folderResults]} {
#		file mkdir $folderResults
#	}
#
#	set folder "$folderResults/$folderTestRunResults"
#	
#	# If we are trying to create a folder for a demo run, then, if it already exists, delete it
#	if {[string match "*Demo*" $folderTestRunResults] && [file exists $folder]} {
#		file delete -force $folder
#		log "Existing demo folder \"$folder\" has been deleted"
#	}
#
#	if {![file exists "$folder"]} {
#		file mkdir "$folder"
#	} else {
#		log "$folder already exists. Aborting."
#		exit
#	}
#}
#------------------------------------------------------------------------------

#
# Public methods start with a lower case letter; private start with upper case
oo::class create transferStep {
	variable catBaseFile	# The file exported from the CAT
	variable catPtpFile		# The file exported from the CAT
	variable cat1ppsFile	# The file exported from the CAT
	variable resultsFile	# The filename of the filtered data file
	variable profileFile	# The filename of the profile to be applied for PTP stimulus
	#variable profileFolder	# The folder in which profiles are stored
	variable captureFile	# The file where the capture data is stored
	variable captureFolder
	variable freq
	variable ampl
	variable sRate
	variable duration
	variable ll
	variable ul
	#variable impairmentProfileFile
	#variable impairmentProfileFolder
	variable stepList
	variable parentSequence
	variable resultsBeforeFilter
	variable resultsAfterFilter
	variable ptpAmplitudeOut
	variable ptpGain
	variable ptpPassFail
	variable 1ppsAmplitudeOut
	variable 1ppsGain
	variable 1ppsPassFail
	
    constructor { tp parent } {		
		set stepList $tp
		set freq [lindex $tp 0]
		set ampl [lindex $tp 1]
		set sRate [lindex $tp 2]
		set duration [lindex $tp 3]
		set ll [lindex $tp 4]
		set ul [lindex $tp 5]
		set profileFile [lindex $tp 6]
		set parentSequence $parent
		set resultsBeforeFilter "UNDEFINED"
		set resultsAfterFilter "UNDEFINED"
		
		#set remote [$parent isRemoteHost]
		#set demo [$parent isDemoMode]
		
		my InitialiseFilesAndFolders
		
		#if {!$remote} {
		#	set profileFolder [pwd]
		#} else {
		#	set profileFolder [parentSequence getRemoteProfileFolder]
		#}
		#set impairmentProfileFile [lindex $tp 6]

		# For remote export from the CAT, we need to remove periods from freq
		#set freqWithUnderbar [string map {"." "_"} $freq] 		
		#set catFile "NoiseTransfer_$freqWithUnderbar.csv"
		
		#set captureFile ""
		#append captureFile "[$parent getTransferType] NoiseTransfer $freq" "Hz.clxz"

		#set captureFolder ""
		#if {!$remote && !$demo} {		
		#	append captureFolder [pwd] "/[$parent getFolderResults]/"
		#} elseif {!$remote && $demo} {
		#	append captureFolder [pwd] "/ReProcess/"
		#} elseif {$remote} {
		#	append captureFolder $remoteCaptureFolder "/"		
		#}
		#XXXDebugTrace "$captureFolder $captureFile"

		
		
		#log "Step created: $dataFile"
	
	}
	
	method print {p} {$p $stepList}
	
	#method getDataFilename {} 		{return $dataFile}
	method getResultsFilename {} 	{return $resultsFile}
	method getFreq {} 				{return $freq}
	method getAmpl {} 				{return $ampl}
	method getDuration {} 			{return $duration}
	method getSampleRate {} 		{return $sRate}
	method getLL {} 				{return $ll}
	method getUL {} 				{return $ul}	
	method getPtpAmplitudeOut {} 	{return $ptpAmplitudeOut}
	method getPtpGain {} 			{return $ptpGain}
	method getPtpPassFail {} 		{return $ptpPassFail}
	method get1ppsAmplitudeOut {} 	{return $1ppsAmplitudeOut}
	method get1ppsGain {} 			{return $1ppsGain}
	method get1ppsPassFail {} 		{return $1ppsPassFail}

	method startStimulus {} {		
		set stimulus [$parentSequence getStimulus]
		set demo [$parentSequence isDemoMode]
		set simulatedData [$parentSequence getSimulatedData]

		XXXDebugTrace "$stimulus"
		if {$stimulus == "PTP"} {
			#set fileAddress $profileFolder "/" $profileFile
			set fileAddress [$parentSequence getFolderProfiles]
			append fileAddress "/" $profileFile
			XXXDebugTrace $fileAddress

			if {!$demo && $simulatedData == "FALSE"} {
				par1 importimpairmentdata 2 $fileAddress

				XXXDebugTrace "Starting impairment...$fileAddress"
				par1 startimpairment
			}
		} elseif {$stimulus == "SYNCE"} {		
			# Convert amplitude from ns to us
			set inputAmplWander [expr {$ampl / 1000}]
	
			par1 paragonset WanderGeneration Tolerance Single Frequency $freq
			par1 paragonset WanderGeneration Tolerance Single Amplitude $inputAmplWander

			par1 paragonset WanderGeneration Tolerance Single Enable START

		} else {
			log "*** ERROR *** startStimulus: Unrecognised stimulus ($stimulus) - must be PTP or SYNCE. Aborting..."
			exit
		}
	}

	method stopStimulus {} {
		set stimulus [$parentSequence getStimulus]
		if { $stimulus == "PTP" } { 
			XXXDebugTrace "Stopping impairment..."
			par1 stopimpairment
		} else {
			XXXDebugTrace "Stopping SyncE wander..."
			# This will stop wander generation at the next 360deg zero crossing
			# Should poll here - might take some time
			par1 paragonset WanderGeneration Tolerance Single Enable FORCESTOP

			set loop  "RUNNING"
			while { $loop != "IDLE" } {
				set loop [ par1 paragonget WanderGeneration Tolerance Single State ]
				#log "."
			}
		}
	}

	
	method doCapture {} {

		set demo [$parentSequence isDemoMode]
		set simulatedData [$parentSequence getSimulatedData]
		set recoveryTime [$parentSequence getRecoveryTime]

		XXXDebugTrace "Starting packet capture - $duration + $recoveryTime seconds"
		
		if {!$demo && $simulatedData == "FALSE"} {
			packetCapture par1 [expr {$duration + $recoveryTime}] $demo
		} elseif {$demo && $simulatedData == "FALSE"} {		
			# Demo mode: load files from the Demo folder
			set f ""
			#append f [pwd] "/Demo/" $captureFile
			append f "$captureFolder/" $captureFile
			if {[file exists $f]} {
				par1 importdata $f
				XXXDebugTrace "Demo Mode - Stopping MSE..."
				#ENT-50 par1 paragonset MasterSlave Master #0 Enabled FALSE
			} else {
				logLevel "FATAL" "$f does not exist"
				exit
			}
		}

		# explicitly stop capture, possible differences in GUI behavious
		# just in case catch, should an older version throw an error on stopping an already stopped capture
		XXXDebugTrace "Stopping packet capture..."
		catch {
			par1 stopcapture
#			puts "capture stopped"
		}
	
		if {!$demo && $simulatedData == "FALSE"} {
			set f ""
			#append f [pwd] "/$folder/" $captureFile
			append f "$captureFolder/" $captureFile
			XXXDebugTrace "Exporting capture to $f"
			if { [catch {par1 exportdata $f}] } {
				log "Failed to export capture file $captureFile (capture status: [par1 paragonget InstrumentStatus Capture IsRunning]). This is not fatal - continuing execution."
			}
		}
	}
	
	# Measurement must be PTP or 1PPS
	method exportFileFromCat {ptp_or_1pps} {
		global P100GNoiseTransferFile

		set simulatedData [$parentSequence getSimulatedData]

		if {$simulatedData == "FALSE"} {
			if {$ptp_or_1pps == "PTP"} {
				set meas "T4"
				set file $catPtpFile
			} elseif {$ptp_or_1pps == "1PPS"} {
				set meas "1PPS"
				set file $cat1ppsFile
			} else {
				log "*** ERROR *** exportFromCat: $ptp_or_1pps is not a valid measurement. Aborting..."
				exit
			}
			XXXDebugTrace "$file, $meas"
		
			if { [catch {par1 paragonset Cat SelectSlot $meas} ] } {
				log "FATAL ERROR (exportFileFromCat): "
				log "\tProblem selecting $meas in CAT"
				log "\tCheck that your instrument and DUT configuration is correct."
				log "\tYou should verify your set up by running a manual test before re-running the script."
				log "Aborting..."
				exit			
			
			}
		
#		set newCatFile $catFile
		# Remove the .csv
#		set newCatFile [ string map {".csv" ""} $catFile ]
#		set catFilePostfix "_$meas"
#		append catFilePostfix "_Cat.csv"
		#puts $catFilePostfix
#		append newCatFile $catFilePostfix
		#set catFile [ string map {".csv" "_Cat.csv"} $catFile ]

			set catFileAddress ""

			set remote [$parentSequence isRemoteHost]

			if {!$remote} {
				# If the script isn't running on a remote host, then provide the full path to the results
				append catFileAddress [pwd] "/" [$parentSequence getFolderResults] "/" $file
			} else {
				# If the script is running on a remote host, then just use the filename
				# In this case, the file will be written to the default location - CAT/Export
				append catFileAddress [file tail $file]
			}
	
	#	
	#	GM Changes
	#	
	#	#par1 paragonset Cat TimeErrorCsv Export $catFileAddress;      # GM: This should be the full data export. Not supported in v18.00

	#	par1 paragonset Cat SelectSlot "2Way"
	#	par1 paragonset Cat TIMEERROR ExportFullCsv $catFileAddress;  # GM: Trace data
	
		#XXXDebugTrace "exportFileFromCat: $catFileAddress"
			par1 paragonset Cat TimeErrorCsv Export $catFileAddress
	
		# Dirty hack - CAT seems to be exporting to its own export folder.
		# Not sure why this isn't taking the full path but hack it for the moment
			if {$P100GNoiseTransferFile} {
				set exportedFile $env(HOME)
				append exportedFile "/Documents/Calnex/Cat/Exports/"
				append exportedFile [file tail $catFileAddress]
				XXXDebugTrace "$exportedFile"
				file rename -force $exportedFile $catFileAddress
			}
		}
	}
	
	method getResults {ptp_or_1pps} {

		if {$ptp_or_1pps == "PTP"} {
			set meas "T4"
			set file $catPtpFile
			set sampleRate 16.0
		} elseif {$ptp_or_1pps == "1PPS"} {
			set meas "1PPS"
			set file $cat1ppsFile
			set sampleRate 1.0
		} else {
			log "*** ERROR *** getResults: $ptp_or_1pps is not a valid measurement. Aborting..."
			exit
		}
		#XXXDebugTrace "getResults: $file, $meas"

		set simulatedData [$parentSequence getSimulatedData]

		if {$simulatedData != "GENERATE"} {
			#set results [readResultsFromFile $catFile $measure $inputFreq]
			set results [my ReadResultsFromFile $ptp_or_1pps]
			XXXDebugTrace "Using data from CAT"
		} else {
			set results [generateSimulatedResults $sRate $freq $ampl $duration]
			XXXDebugTrace "Using simulated data"
		}
			
		# Remove a specified amount of data from the capture prior to filtering
		# This allows the device to recover after being hit by the noise
		# The recovery time should be set as a parameter
		#set results [removeRecoveryTime $results $deviceRecoveryTime $inputSampleRate $inputFreq $measure]
		set resultsBeforeFilter [removeRecoveryTime $results [$parentSequence getRecoveryTime] $sampleRate $freq $meas]
	}
	
	# Private
	method ReadResultsFromFile {ptp_or_1pps} {
		global guiHostname

		set remote [$parentSequence isRemoteHost]
		set measure [$parentSequence getMeasurement]
		set resultsFolder [$parentSequence getFolderResults]
		set simulatedData [$parentSequence getSimulatedData]

		if {$ptp_or_1pps == "PTP"} {
			set meas "T4"
			set file $catPtpFile
		} elseif {$ptp_or_1pps == "1PPS"} {
			set meas "1PPS"
			set file $cat1ppsFile
		} else {
			logLevel "FATAL" "$ptp_or_1pps is not a valid measurement"
			exit
		}
		XXXDebugTrace "$file, $meas"
		
		set fileAddress ""
		append fileAddress $resultsFolder "/$file"
		
		if {$simulatedData != "FALSE" && $simulatedData != "GENERATE" && $simulatedData != "TRUE"} {
			# We are evaluating noise performance
			# Load a file that has the noise type included just before the file extension
			set filenameNoExt [file rootname $file]
			set fileExt [file extension $file]
			set newFilename ""
			append newFilename $filenameNoExt "_" $simulatedData $fileExt
			XXXDebugTrace "Simulated data - $newFilename, $meas"
			set fileAddress ""
			append fileAddress "[pwd]/ReProcess" "/$newFilename"
			
		}
		
		#XXXDebugTrace "ReadResultsFromFile: remoteHost $remote, $fileAddress"
	
		if {!$remote} {
			return [readResultsFromLocalFile $fileAddress $ptp_or_1pps $freq]
		} else {
			# The P-X GUI and CAT are hosted on a remote PC
			# The file has been exported on this PC to the CAT/Export folder - CHECK THIS
			# Use the CAT REST API to get the contents of the file
			# THis will be returned in the same format as the csv and so we should process it the same way
			XXXDebugTrace "Fetching results from remote PC..."
			set restApi "http://$guiHostname:1337/api/"
			#puts [file tail $catFile]
			set config "method {get} headers \{Content-Type text/csv\}"
			set url "cat/filecommander/download/ExportFolder?AsAttachment=true&FileId=[file tail $fileAddress]"

			set f ""
			append f [pwd] "/$fileAddress"
			#puts $f
			set fid [open $f "w"]
			puts $fid [rest::simple "${restApi}${url}" "" $config ""]
			close $fid
		
			# Now do the same as we normally do...
			return [readResultsFromLocalFile $fileAddress $ptp_or_1pps $freq]
		}
	}
	

	method applyFilter {ptp_or_1pps} {

		if {$resultsBeforeFilter == "UNDEFINED"} {
			log "resultsBeforeFilter not defined. Aborting..."
			exit
		}

		set filter [$parentSequence getFilterType]
		set stim [$parentSequence getStimulus]
		set resultsFolder [$parentSequence getFolderResults]

		set resultsPath "$resultsFolder/$resultsFile"

		XXXDebugTrace "$stim $filter $resultsFile $ptp_or_1pps"
		if {$filter == "LSQ"} {
			if {$ptp_or_1pps == "PTP"} {
				set resultsAfterFilter [filterLeastSquares $stim $resultsBeforeFilter $resultsPath $freq $ampl $sRate]
			} else {
				set resultsAfterFilter [filterLeastSquares $stim $resultsBeforeFilter $resultsPath $freq $ampl 1.0]
			}
		} elseif {$filter == "IIR"} {
			if {$ptp_or_1pps == "PTP"} {
				# Get the sample rate - this is needed for the IIR filter
				# We assume that the sample rate is the equivalent of the Sync rate. For MSE BC testing, this is probably reasonable
				par1 addMeasurement "measTE2Way" "TE" "2Way"
				set measTE2Way [ par1 getMeasurement "measTE2Way" ]
				set measTE2WayResult [ $measTE2Way createResult ]
				set inputSampleRate [ $measTE2WayResult getForwardRate ]
				#log $max_min
				set resultsAfterFilter [filterIIR $stim $file $resultsBeforeFilter $resultsFile $freq $ampl $sRate]
			} else {
				log "ERROR: Unsupported filter for 1pps. Filter type: $filter. Aborting..."
				exit
			} 
		} elseif {$filter == "NONE"} {
			if {$ptp_or_1pps == "PTP"} {
				# No filtering so just get the max-min result for T4 from the CAT
				par1 addMeasurement "measTET4" "TE" "T4"
				set measTET4 [ par1 getMeasurement "measTET4" ]
				set measTET4Result [ $measTET4 createResult ]
				set max_min [ $measTET4Result getMax_Min ]
				#log $max_min
				# Create the result list (bit artifical but it maintains the filter interface)
				set resultsAfterFilter [list $ampl $max_min $freq]
			} else {
				# No filtering so just get the max-min result for 1pps from the CAT
				set y [lindex $resultsBeforeFilter 1]
				# Result is the range of the data
				#puts [getMax $y]
				#puts [getMin $y]
				set max_min [expr {[getMax $y] - [getMin $y]}]
				#log $max_min
				# Create the result list (bit artifical but it maintains the filter interface)
				set resultsAfterFilter [list $ampl $max_min $freq]
			}
		} else {
			log "ERROR: Unrecognised filter type: $filter. Aborting..."
			exit
		}

	}

	
	method calculateFinalResults {ptp_or_1pps} {
		#XXXDebugTrace "calculateFinalResults: $ptp_or_1pps"
		if {$resultsAfterFilter == "UNDEFINED"} {
			log "resultsAfterFilter not defined. Aborting..."
			exit
		}

		set measurement [$parentSequence getMeasurement]

		if {$ptp_or_1pps == "PTP"} {
			#set pkpkAmplitudeIn [ lindex $resultsAfterFilter 0] 
			set ptpAmplitudeOut  [ lindex $resultsAfterFilter 1] 
			#set frequency [ lindex $resultsAfterFilter 2]
			#XXXDebugTrace "calculateFinalResults: $ptpAmplitudeOut"
	    
			# Calculate the gain
			#set ptpGain [ expr 20 * log10( [expr $ptpAmplitudeOut / $pkpkAmplitudeIn ])]
			if {$ptpAmplitudeOut == 0} {
				set ptpGain "-100.00"
			} else {
				set ptpGain [ expr {20 * log10($ptpAmplitudeOut / $ampl)}]				
			}
			#XXXDebugTrace "calculateFinalResults: $ptpGain"
		
			# Work out pass/fail
			# There is one pass/fail column: pass, fail (lower), fail (upper)
			set ptpPassFail [my DeterminePassFail $ptpAmplitudeOut]
			#XXXDebugTrace "calculateFinalResults: $ptpPassFail"
		
			# save results to file 
#			if {$measurement != "BOTH"} {
#				#savePtpResultsToFile $openedFileResultsGeneral $file $frequency $pkpkAmplitudeIn $ptpAmplitudeOut $ptpGain $lowerLimit $upperLimit $ptpPassFail
#				savePtpResultsToFile [sequence getFidResults] $resultsFile $freq $ampl $ptpAmplitudeOut $ptpGain $ll $ul $ptpPassFail
#			}
		} elseif {$ptp_or_1pps == "1PPS"} {
			#set pkpkAmplitudeIn [ lindex $resultList 0] 
			set 1ppsAmplitudeOut  [ lindex $resultsAfterFilter 1] 
			#set frequency [ lindex $resultList 2]
			#set pkpkAmplitudeIn $inputAmpl
			#set frequency $inputFreq
			#set amplitudeOut  $max_min

			#puts " Before gain calc AmplitudeIn $pkpkAmplitudeIn amplitudeOut  $amplitudeOut  frequency $frequency " 

			# calculate gain
			if {$1ppsAmplitudeOut == 0} {
				set 1ppsGain "-100.00"
			} else {
				set 1ppsGain [ expr {20 * log10($1ppsAmplitudeOut / $ampl)}]
			}
			#puts "After gain calc AmplitudeIn $pkpkAmplitudeIn amplitudeOut  $amplitudeOut  frequency $frequency gain $gain" 

			# Work out pass/fail
			# There is one pass/fail column: pass, fail (lower), fail (upper)
			set 1ppsPassFail [my DeterminePassFail $1ppsAmplitudeOut]
		
			# save results to file 
#			if {$measurement != "BOTH"} {
				#save1ppsResultsToFile $openedFileResultsGeneral $file $frequency $pkpkAmplitudeIn $1ppsAmplitudeOut $1ppsGain $lowerLimit $upperLimit $1ppsPassFail
#				save1ppsResultsToFile [sequence getFidResults] $resultsFile $freq $ampl $1ppsAmplitudeOut $1ppsGain $ll $ul $1ppsPassFail
#			}
		
		} else {
			log "calculateFinalResults: Argument ($ptp_or_1pps) not supported. Aborting..."
			exit
		}
	
	}
	
	method saveAndDisplayFinalResults {meas} {
		#XXXDebugTrace "saveAndDisplayFinalResults: $ptp_or_1pps"
		if {$meas == "PTP"} {
			savePtpResultsToFile [sequence getFidResults] $resultsFile $freq $ampl $ptpAmplitudeOut $ptpGain $ll $ul $ptpPassFail
		} elseif {$meas == "1PPS"} {
			save1ppsResultsToFile [sequence getFidResults] $resultsFile $freq $ampl $1ppsAmplitudeOut $1ppsGain $ll $ul $1ppsPassFail
		} elseif {$meas == "BOTH"} {
			saveResultsToFile [sequence getFidResults] $resultsFile $freq $ampl $ptpAmplitudeOut $ptpGain $1ppsAmplitudeOut $1ppsGain $ll $ul $ptpPassFail $1ppsPassFail		
		} else {
			log "calculateFinalResults: Argument ($meas) not supported. Aborting..."
			exit
		}

	}
	
	
	
	
	# Private
	method DeterminePassFail {in} {
		if {$ll != "NA" && $in < $ll} {return "FAIL(lower)"}
		if {$ul != "NA" && $in > $ul} {return "FAIL(upper)"}
		return PASS
	}

	# Private
	method InitialiseFilesAndFolders {} {
		set remote [$parentSequence isRemoteHost]
		set demo [$parentSequence isDemoMode]

#		if {!$remote} {
#			set profileFolder [pwd]
#		} else {
#			set profileFolder [$parentSequence getRemoteProfileFolder]
#		}

		# For remote export from the CAT, we need to remove periods from freq
		set freqWithUnderbar [string map {"." "_"} $freq] 		
		set catBaseFile "NoiseTransfer_$freqWithUnderbar.csv"
		
		set resultsFile $catBaseFile
		set resultsFile [string map {".csv" "_Results.csv"} $resultsFile] 

		# Create filenames of each possible measurement
		# Remove the .csv
		set catPtpFile [ string map {".csv" ""} $catBaseFile ]
		set catFilePostfix "_T4"
		append catFilePostfix "_Cat.csv"
		append catPtpFile $catFilePostfix

		set cat1ppsFile [ string map {".csv" ""} $catBaseFile ]
		set catFilePostfix "_1PPS"
		append catFilePostfix "_Cat.csv"
		append cat1ppsFile $catFilePostfix
		
		
		set captureFile ""
		append captureFile "[$parentSequence getTransferType] NoiseTransfer $freq" "Hz.clxz"

		set captureFolder ""
		if {!$remote && !$demo} {		
			append captureFolder [pwd] "/[$parentSequence getFolderResults]/"
		} elseif {!$remote && $demo} {
			append captureFolder [pwd] "/ReProcess/"
		} elseif {$remote} {
			#append captureFolder $remoteCaptureFolder "/"		
			append captureFolder [$parentSequence getRemoteCaptureFolder]
		}
		#XXXDebugTrace "$captureFolder $captureFile"

	}
}

oo::class create transferSequence {
	variable sequence {}
	variable folderProfiles
	variable folderResults "Results"
	variable folderTestRunResults "UNDEFINED"
	variable fileOverallResults "UNDEFINED"
	variable fidOverallResults
	variable transferType "UNDEFINED"
	variable stimulus "UNDEFINED"
	variable measurement "UNDEFINED"
	variable parametersDictionary
	variable dateStart timeStart
	variable stepList
	variable remoteHost False
	variable remoteCaptureFolder "UNDEFINED"
	variable remoteProfileFolder "UNDEFINED"
	variable demoMode
	variable P100G_ptpCapture
	variable P100G_1ppsCapture
	variable recoveryTime
	variable guardTime
	
    constructor {noiseTransferType params demo remote} {
		#XXXDebugTrace "transferSequence: creating... "
		set parametersDictionary $params
		set remoteHost $remote
		set demoMode $demo
		
		set systemTime [clock seconds]
		set dateStart [clock format $systemTime -format "%Y-%m-%d"]
		set timeStart [clock format $systemTime -format %H-%M-%S]

		set transferType $noiseTransferType

		set recoveryTime [dict get $parametersDictionary "RecoveryTime"]
		set guardTime "UNDEFINED"
		
		my InitialiseFilesAndFolders
		
#		# Profiles folder
#		if {!$remote} {
#			set folderProfiles [pwd]
#		} else {
#			set folderProfiles [dict get $parametersDictionary "RemoteProfileFolder"]
#		}

#		# Results folder and sub-folders
#		set folderResults "Results"		

#		if {!$demoMode} {
#			set folderTestRunResults "Results $noiseTransferType $dateStart-T-$timeStart"
#		} else {
#			set folderTestRunResults "Results $noiseTransferType Demo"
#		}
#		set fileOverallResults "$folderResults/$folderTestRunResults/Results $dateStart-T-$timeStart.csv"

#		createResultsFolders $folderResults $folderTestRunResults

		log "\nStart: $dateStart $timeStart\n"
#		log "Results will be stored in:\t\"$folderResults\\$folderTestRunResults\""
		
		if {[string match "PTP-*" $noiseTransferType]} {
			set stimulus "PTP"
		} elseif {[string match "SyncE-*" $noiseTransferType]} {
			set stimulus "SYNCE"
		} 

		if {[string match "*-PTP" $noiseTransferType]} {
			set measurement "PTP"
		} elseif {[string match "*-1PPS" $noiseTransferType]} {
			set measurement "1PPS"
		} elseif {[string match "*-BOTH" $noiseTransferType]} {
			set measurement "BOTH"
		}

#		set fidOverallResults [ open $fileOverallResults "a"]
#		my WriteHeaderToResultsFile
	}
	destructor {
		close $fidOverallResults
	}
	
	method loadStepList {sList} {
		set stepList $sList
		set counter 0
		foreach tp $sList {
			set s [transferStep create step$counter $tp [self object]]
			#puts $s
			my addStep $counter $s
			incr counter
		}
	}
	method getStepList {} {return $stepList}
	
	method outputStepListAsTcl {} {
		#set sListFilename "stimulusList.tcl"
		set sListFilename "stimulusList_$stimulus.tcl"
		set sListPath "$folderResults/$folderTestRunResults/$sListFilename"
		
		# If the file already exists, delete it
		if {[file exists $sListPath]} {
			file delete -force $sListPath
			log "Existing stimulus list file \"$sListPath\" has been deleted"
		}

		set fid [open $sListPath "a"]
		
		puts $fid "# THIS FILE HAS BEEN GENERATED BY THE G.8273.2 SCRIPT. DO NOT DELETE OR MODIFY"
		puts $fid "set stimulusListUsedDuringTest \{ "
		foreach tp $stepList {
			puts -nonewline $fid "\{ "
			puts -nonewline $fid "[lindex $tp 0] "
			puts -nonewline $fid "[lindex $tp 1] "
			puts -nonewline $fid "[lindex $tp 2] "
			puts -nonewline $fid "[lindex $tp 3] "
			puts -nonewline $fid "[lindex $tp 4] "
			puts -nonewline $fid "[lindex $tp 5] "
			puts $fid "[lindex $tp 6] \}"
		}
		puts $fid " \}"
		flush $fid
		close $fid
	}
	
	method addStep {index step} {
		#puts "adding step: $step"
		set s [list $index $step]
		lappend sequence $s
		#puts $sequence
	}
	method getSequence {} {return $sequence}
	
	method print {p} {
		foreach s $sequence {
			set i [lindex $s 0]
			set step [lindex $s 1]
			$step print $p
		}
	}
	
	method p100G_setPtpCapture {c} 	{set P100G_ptpCapture $c}
	method p100G_getPtpCapture {} 	{return $P100G_ptpCapture}
	method p100G_set1ppsCapture {c} {set P100G_1ppsCapture $c}
	method p100G_get1ppsCapture {} 	{return $P100G_1ppsCapture}
	
	method getTransferType {} 		{return $transferType}
	method getStimulus {} 			{return $stimulus}
	method getMeasurement {} 		{return $measurement}
	method getStartTime {} 			{return $timeStart}
	method getStartDate {} 			{return $dateStart}
	method getSettlingTime {} 		{return [dict get $parametersDictionary "SettlingTime"]}
	method setRecoveryTime {t} 		{set recoveryTime $t}
	method getRecoveryTime {} 		{return $recoveryTime}
	method setGuardTime {t} 		{set guardTime $t}
	method getGuardTime {} 			{return $guardTime}
	method getFilterType {} 		{return [dict get $parametersDictionary "Filter"]}
	method getFidResults {} 		{return $fidOverallResults}
	method getFolderResults {} 		{return "$folderResults/$folderTestRunResults"}
	
	method printParameters {} {
		log "Transfer Sequence Parameters:"
		dict for {key value} $parametersDictionary {
			# DJWAT AE-7
			if { $key == "LineRate" || $key == "Interface" || $key == "AutoNegotiate"} {
				log "\tPort - $key"
				
			} elseif { $key == "Port1Config" || $key == "Port2Config" } {
			} else {
				log "\t$key - $value"
			}
		}
	}
	method getDataFilename {} {
		#		return $dataFile
	}
	method getResultsFilename {} {
#		return $resultsFile
	}
	method isRemoteHost {} 				{return $remoteHost}
	method getRemoteCaptureFolder {} 	{return [dict get $parametersDictionary "RemoteCaptureFolder"]}
	method getSimulatedData {} 			{return [dict get $parametersDictionary "SimulatedData"]}
	method getFolderProfiles {} 		{return $folderProfiles}
	method isDemoMode {} 				{return $demoMode}
	
	# Private
	method WriteHeaderToResultsFile {} {
		#set systemTime [clock seconds]
		#set d [clock format $systemTime -format "%d %b %Y"]
		#set t [clock format $systemTime -format %H:%M:%S]
		set resultsInfo ""
		if {$transferType == "PTP-BOTH" || $transferType == "SyncE-BOTH"} {
			append resultsInfo "\nNoise Transfer ($transferType)" "," "$dateStart $timeStart\n" Profile "," "Freq (Hz)" "," "Amplitude (p-p ns)" "," "PTP Measured (p-p ns)" "," "PTP Gain (dB)" "," "1pps Measured (p-p ns)" "," "1pps Gain (dB)" "," "Lower Limit (p-p ns)" "," "Upper Limit (p-p ns)" "," "Pass/Fail"
		} elseif {$transferType == "PTP-PTP" || $transferType == "SyncE-PTP"} {
			append resultsInfo "\nNoise Transfer ($transferType)" "," "$dateStart $timeStart\n" Profile "," "Freq (Hz)" "," "Amplitude (p-p ns)" "," "PTP Measured (p-p ns)" "," "PTP Gain (dB)" "," "Lower Limit (p-p ns)" "," "Upper Limit (p-p ns)" "," "Pass/Fail"
		} elseif {$transferType == "PTP-1PPS" || $transferType == "SyncE-1PPS"} {
			append resultsInfo "\nNoise Transfer ($transferType)" "," "$dateStart $timeStart\n" Profile "," "Freq (Hz)" "," "Amplitude (p-p ns)" "," "1pps Measured (p-p ns)" "," "1pps Gain (dB)" "," "Lower Limit (p-p ns)" "," "Upper Limit (p-p ns)" "," "Pass/Fail"
		} else {
			log "WriteHeaderToResultsFile: Unknown Type - $transferType. Aborting..."
			exit
		}
		puts $fidOverallResults $resultsInfo

	}
	# Private
	method InitialiseFilesAndFolders {} {
		# Profiles folder
		if {!$remoteHost} {
			set folderProfiles [pwd]
		} else {
			set folderProfiles [dict get $parametersDictionary "RemoteProfileFolder"]
		}

		# Results folder and sub-folders
		set folderResults "Results"		

		if {!$demoMode} {
			set folderTestRunResults "Results $transferType $dateStart-T-$timeStart"
		} else {
			set folderTestRunResults "Results $transferType Demo"
		}
		set fileOverallResults "$folderResults/$folderTestRunResults/Results $dateStart-T-$timeStart.csv"

		#createResultsFolders $folderResults $folderTestRunResults
		if {![file exists $folderResults]} {
			file mkdir $folderResults
		}

		set folder "$folderResults/$folderTestRunResults"
	
		# If we are trying to create a folder for a demo run, then, if it already exists, delete it
		if {[string match "*Demo*" $folderTestRunResults] && [file exists $folder]} {
			file delete -force $folder
			log "Existing demo folder \"$folder\" has been deleted"
		}

		if {![file exists "$folder"]} {
			file mkdir "$folder"
		} else {
			log "$folder already exists. Aborting."
			exit
		}

		set fidOverallResults [ open $fileOverallResults "a"]
		my WriteHeaderToResultsFile

		log "Results will be stored in:\t\"$folderResults\\$folderTestRunResults\""
	}
}

