## Based on code from Gaby Zigelboim (RAD)
##
## Re-factored to separate input data from least squares algorithm and gain
## calculation
##
##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
##%% Assumptions:
##%% - Sample frequency is fixed at 16 samples/sec based on G.8275.1
##%%   Note that this means that this algorithm can ONLY be used under
##%%   script/instrument control
##%%   Questions: Can we get the CAT to check this? 
##%%              Does the capture file contain the packet rate so that the script
##%%              can check?
##%% - The algorithm will be called for each step in the noise transfer table.
##%%   The input frequency and amplitude for each step must be input to the algorithm
##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#set pi 3.1415926535897931

#proc filterLeastSquares { stimulus  testFile catFile resultsFile inputFreq inputAmpl inputSampleRate } { 
proc filterLeastSquares { stimulus results resultsFile inputFreq inputAmpl inputSampleRate } { 
	variable pi
	
	XXXDebugTrace "Using Least Squares filter... $inputFreq $inputAmpl $inputSampleRate"
	
	# open file to write
	set fileId [ open $resultsFile "a"]

	append intro "t1" "," "d1" "," 
	puts $fileId $intro


    set inputAmplitude $inputAmpl 
    set f $inputFreq
    set Fs $inputSampleRate 

	set t [lindex $results 0]
	set y [lindex $results 1]
	#array set y $results
	# y now contains the the measured amplitude in ns for each sampled point
	
	# Set up filter params
	#XXXDebugTrace "filterLeastSquares: Initialising..."
	set sumCosCos 0
	set sumCosSin 0
	set sumSinSin 0
	set sumCosY	  0
	set sumSinY	  0
	set sumYY	  0

	set N [llength $y]
	#set N [array size y]
	XXXDebugTrace "filterLeastSquares: Input data max: [getMax $y] min: [getMin $y] range: [expr {[getMax $y] - [getMin $y]}]"
	XXXDebugTrace "filterLeastSquares: $N samples"
	
	set offset [expr {[getMax $y] + [getMin $y]}]
	set offset [expr {$offset / 2}]
	#set y [expr $y1 - $o]
	#set y $y1
	XXXDebugTrace "Offset: $offset"
	# (>3) minimal number of samples before starting the estimation
	#set nThresh	10

	#%% Calculate the sums for all points in the data
	#%% Note that the original alogorithm terminated early if the estimate at any
	#%% time was less than limit-3*std. We can't do this - we need to produce an actual
	#%% gain value not simply verify that it is less than the limit. 
	#%% Likewise, the original algorithm terminated if the amplitude and deviation
	#%% estimate were clearly above the limit. Again, we need an actual gain value.
	#%% Note:
	#%% - If ampEst-(sigmaEst*3)>Limit, this means that the algorithm is failing to converge
	#%%   However, we ignore this and finish since we need a result
	#%% - If sigmaEst*10 < ampEst then we have a valid result otherwise the SNR is not good
	#%%   In the latter case, it likely that the algortihm will not converge and so the result is questionable.
	#%%   The assumption is that, if this is the case, we will fail limits.
	#XXXDebugTrace "filterLeastSquares: Estimating..."
	for {set n 1}  {$n < $N} {incr n} {
#	for n=1:N	
		set i [expr {$n-1}]
		set d [lindex $y $n]
		set d [expr {$d - $offset}]
		set tn [lindex $t $n]
		#XXXDebugTrace $d
		#set cos_		[expr cos(2*$pi*$f*($n/$Fs))]
		#set sin_		[expr sin(2*$pi*$f*($n/$Fs))]
		set cos_		[expr {cos(2*$pi*$f*$tn)}]
		set sin_		[expr {sin(2*$pi*$f*$tn)}]
		set sumCosCos	[expr {$sumCosCos + $cos_ * $cos_}]
		set sumCosSin	[expr {$sumCosSin + $cos_ * $sin_}]
		set sumSinSin	[expr {$sumSinSin + $sin_ * $sin_}]
		#set sumCosY		[expr $sumCosY	 + $cos_ * $y($i)]
		#set sumSinY		[expr $sumSinY	 + $sin_ * $y($i)]
		#set sumYY		[expr $sumYY	 + $y($i) * $y($i)]
		set sumCosY		[expr {$sumCosY	 + $cos_ * $d}]
		set sumSinY		[expr {$sumSinY	 + $sin_ * $d}]
		set sumYY		[expr {$sumYY	 + $d * $d}]

		set info "" 
		#append info $i "," $y($i) "," 
		append info $i "," $d "," 
		# puts $info
		puts $fileId $info
	}

	set detMat		[expr {$sumCosCos*$sumSinSin - $sumCosSin*$sumCosSin}]
	set invMat11	[expr {$sumSinSin / $detMat}]
	set invMat12	[expr {-$sumCosSin / $detMat}] 
	set invMat21	[expr {-$sumCosSin / $detMat}]
	set invMat22	[expr {$sumCosCos / $detMat}]
	set aCos		[expr {$invMat11 * $sumCosY	+ $invMat12 * $sumSinY}]
	set aSin		[expr {$invMat21 * $sumCosY	+ $invMat22 * $sumSinY}]
	set sigmaSquare	[expr {($sumYY - ($aCos * $sumCosY + $aSin * $sumSinY))/$N}]

	set ampEst		[expr {sqrt($aCos*$aCos + $aSin*$aSin)}]
	set sigmaEst	[expr {sqrt($sigmaSquare/$N)}]

    set amplitude   $ampEst

	if {$sigmaEst*10 > $ampEst} {
		# SNR is low. Flag this only when debugging
		XXXDebugTrace "!!!!!!!!!!!!!!  Questionable result  !!!!!!!!!!!!!!!!!!!!!!!!!"
	}

	set amplitude [expr {2*$amplitude}]
	XXXDebugTrace "LSQ: Input Amplitude: $inputAmplitude, Output Amplitude:  $amplitude"
	close $fileId
	
	set resultList [ list $inputAmplitude $amplitude $inputFreq ]
	return $resultList
}


	
##### Octave code.... #############
# global Fs - sample frequency
#
#%% Input Parameters - assume one point for the moment
#f	=	0.245;			% Input tone for noise transfer testing
#amp	=	200;			% Input tone amplitude
#
#
#%% Simulation Parameters - used ONLY for simulation.
#%% These are NOT required for the final algorithm
#T		    =	100;	% Maximal simulation time (seconds)
#gainBC	=	0.05;	% The simulated gain of the T-BC
#
#%% Limit not used in this version of the alogirthm
#%limit_dB	=	0.1;	% [dB] limit of output amplitude
#
#%% Algorithm constants - this is global because I am being lazy
#Fs		=	16;		% Sampling Frequency (constant in G.8275.1)
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Test data generation
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% This generates data to use as simulated input. 
#%% In real-life, this data will be the 2Way Time Error measured by the CAT
#%%
#%% Input Parameters
#%% - Fsample: 	sample rate
#%% - freq:      Frequency of the generated data
#%% - amplitude: Amplitude of the generated data
#%% - duration: 	period over which the data is generated
#%% Output
#%% - testData: array of amplitude values
#function testData = generateTestData (Fsample, freq, amplitude, duration)
#  sigma	=	6;		% Gaussian noise STD
#	N = Fsample * duration;
#	t = (1:N)'/Fsample;
#	x =	amplitude*sin(2*pi*freq*t+0.12345678);	% Add a constant "unknown" phase for repeatability of the simulation
#	u =	randn(size(x))*sigma;
#	y =	x+u;
#	testData = y;
#endfunction
#
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Algorithm
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Input Parameters
#%% - y:     array of variable size containing the measured amplitude in ns for 
#%%          each sampled point
#%% - f:     the frequency of the input noise applied to the DUT
#%% - amp:   the amplitude of the input noise applied to the DUT
#%% Output
#%% - pkpkAmplitude: pkpk amplitude of the smoothed / filtered data
#function amplitude=leastSquaresEstimate (y, f, amp)
#  global Fs;
#  %% Run Estimation
#  sumCosCos	=	0;
#  sumCosSin	=	0;
#  sumSinSin	=	0;
#  sumCosY		=	0;
#  sumSinY		=	0;
#  sumYY		=	0;

#  N = length(y);
#  %nThresh	=	10;		% (>3) minimal number of samples before starting the estimation

#  %% Calculate the sums for all points in the data
#  %% Note that the original alogorithm terminated early if the estimate at any
#  %% time was less than limit-3*std. We can't do this - we need to produce an actual
#  %% gain value not simply verify that it is less than the limit. 
#  %% Likewise, the original algorithm terminated if the amplitude and deviation
#  %% estimate were clearly above the limit. Again, we need an actual gain value.
#  %% Note: Should check with Gaby: 
#  %% - if ampEst-(sigmaEst*3)>Limit, does this mean
#  %%   that the algorithm is failing and that we will not get a sensible result?
#  %% - Should we use sigmaEst*10 < ampEst as a check for a valid result?
#  %%   Is it likely that the algortihm will not converge and the result is not sensible?
#  %%   What do we do if we don't have a valid result?
#  for n=1:N
#	  cos_		  =	cos(2*pi*f*n/Fs);
#	  sin_		  =	sin(2*pi*f*n/Fs);
#	  sumCosCos	=	sumCosCos	+ cos_ * cos_;
#	  sumCosSin	=	sumCosSin	+ cos_ * sin_;
#	  sumSinSin	=	sumSinSin	+ sin_ * sin_;
# 	sumCosY		=	sumCosY		+ cos_ * y(n);
#	  sumSinY		=	sumSinY		+ sin_ * y(n);
#	  sumYY		  =	sumYY		  + y(n) * y(n);
#  end		% for n=1:N
	
#%  	if n>nThresh
#  		detMat		=	sumCosCos*sumSinSin - sumCosSin*sumCosSin;
#  		invMat11	=	 sumSinSin / detMat;	% the devision can be done later as well
#  		invMat12	=	-sumCosSin / detMat;
#  		invMat21	=	-sumCosSin / detMat;
#  		invMat22	=	 sumCosCos / detMat;
#  		aCos		=	invMat11 * sumCosY	+ invMat12 * sumSinY;
#  		aSin		=	invMat21 * sumCosY	+ invMat22 * sumSinY;
#  		%sigmaSquare		=	(sumYY - (aCos * sumCosY + aSin * sumSinY))/n;
#  		sigmaSquare		=	(sumYY - (aCos * sumCosY + aSin * sumSinY))/N;

		
#  		ampEst			=	sqrt(aCos*aCos + aSin*aSin);
#  		%sigmaEst		=	sqrt(sigmaSquare/n);
#  		sigmaEst		=	sqrt(sigmaSquare/N);

#      amplitude = ampEst;
      
#%  		flag_STOP_OK		=	0;
#%  		flag_COMPIANT		=	0;
#%  		flag_NON_COMPIANT	=	0;
#%  		if ampEst + sigmaEst*3 < limit
#%  			flag_COMPIANT		=	1;
#%  			flag_STOP_OK		=	1;
#%  		end		% if ampEst + sigmaEst*3 < limit
#%  		if ampEst - sigmaEst*3 > limit
#%  			flag_NON_COMPIANT	=	1;
#%  			flag_STOP_OK		=	1;
#%  		end		% if ampEst - sigmaEst*3 > limit
		
#%  		if n>=Fs*30 && flag_STOP_OK && sigmaEst*10 < ampEst
#	  		% gainEst_dB is good enough.
#		  	% Waiting for 30 is not mandatory, but this is a simulation for Gaussian.
#			  % In real life the noise will be uniform + wander, so best wait at least 30 seconds before
#			  % stopping.
#%			  break;
#%  		end		% if flag_STOP_OK && sigmaEst*10 < ampEst
#%    end		% if n>nThresh
#endfunction

#ampOut=	10^(gainBC/20)*amp;
#testData = generateTestData (Fs, f, ampOut, T);
#ampEst = leastSquaresEstimate (testData, f, amp)
#gainEst_dB		=	20*log10(ampEst/amp)
